#!/usr/bin/env python


import os, sys, os.path, re, subprocess

topology_exec = "/home/hpc/unrz/unrz139/Apps-arm/bin/likwid-topology"
topology_re_size = re.compile("^Size:\s+(.*)")
re_size_unit = re.compile("(\d+)\s(\w+)")

cachesizes = []

def get_caches():
    level = 0
    print("here")
    p = subprocess.Popen(topology_exec, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    p.wait()
    sout = p.stdout.read()
    if p.returncode != 0:
        print("err", p.returncode )
        return level
    for line in sout.split("\n"):
        if line.startswith("Size:"):
            print(line)
            string = topology_re_size.match(line).group(1).strip()
            size, unit = re_size_unit.match(string).groups()
            if unit == "kB":
                size = int(size)*1024
            elif unit == "MB":
                size = int(size)*1024*1024
            cachesizes.append(size)
            level += 1
    fp = open("/proc/meminfo")
    f = fp.read().strip().split("\n")
    fp.close()
    for line in f:
        if line.startswith("MemTotal:"):
            linelist = re.split("\s+", line)
            size = int(linelist[1])
            if linelist[2] == "kB":
                size *= 1024
            elif linelist[2] == "MB":
                size *= 1024*1024
            if size > 1024*1024*1024:
                size = 1024*1024*1024
            cachesizes.append(size)
    return level

def get_important_tests():
    important = ["L2", "L3", "MEM", "CLOCK", "UOPS", "HA"]
    adjust = []
    regular = []
    fp = open("SET.txt")
    f = fp.read().strip().split("\n")
    fp.close()
    for line in f:
        found = False
        for imp in important:
            if imp in line:
                adjust.append(line)
                found = True
        if not found:
            regular.append(line)
    return adjust, regular

def adjust_tests(testgroup):
    fp = open("TESTS/"+testgroup+".txt", "r")
    f = fp.read().strip().split("\n")
    fp.close()
    newdata = []
    level = re.match("L(\d+)", testgroup)
    if level:
        level = int(level.group(1))-1
    else:
        level = len(cachesizes)-1
    print(level, cachesizes)
    min_size = int((cachesizes[level-1] + (0.3*cachesizes[level-1]))/1024)
    max_size = int((cachesizes[level] - (0.2*cachesizes[level]))/1024)
    diff = (cachesizes[level] - cachesizes[level-1])/1024
    step = diff/5
    i = 0
    while i < len(f):
        if not f[i].startswith("VARIANT"):
            newdata.append(f[i]+"\n")
            i+=1
        else:
            count = 0
            for j in range(i,i+4):
                if f[j].startswith("VARIANT"):
                    count += 1
                else: break
            i += count
            newdata.append("VARIANT %dkB 1000\n" % (int(min_size+step),))
            newdata.append("VARIANT %dkB 1000\n" % (int(min_size+(2*step)),))
            newdata.append("VARIANT %dkB 1000\n" % (int(min_size+(3*step)),))
            newdata.append("VARIANT %dkB 1000\n" % (int(min_size+(4*step)),))
    fp = open("TESTS/"+testgroup+".txt", "w")
    for line in newdata:
        fp.write(line)
    fp.close()

level = get_caches()
adjust, regular = get_important_tests()
for testgroup in adjust:
    print("Adjusting "+testgroup)
    adjust_tests(testgroup)
if len(regular) > 0:
    print("Not adjusting:")
    print(regular)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# =======================================================================================
#
#      Filename:  check_data_files.py
#
#      Description:  Basic checks for performance event list files
#
#      Author:   Mikhail Terekhov, termim@gmail.com
#      Project:  likwid
#
#      Copyright (C) 2016 RRZE, University Erlangen-Nuremberg
#
#      This program is free software: you can redistribute it and/or modify it under
#      the terms of the GNU General Public License as published by the Free Software
#      Foundation, either version 3 of the License, or (at your option) any later
#      version.
#
#      This program is distributed in the hope that it will be useful, but WITHOUT ANY
#      WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
#      PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License along with
#      this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =======================================================================================

from collections import OrderedDict
import json
import logging as log
from pathlib import Path
from pprint import pprint, pformat
import sys

from pyparsing import *
ParserElement.setDefaultWhitespaceChars(' \t')



def err(fmt, *args, **kw):
    log.error(fmt.format(*args, **kw))



class EventParser():
    """
    The expected event list EBNF gramma is:

    eol = '\n' ;
    letter = 'A' - 'Z' | 'a' - 'z' ;
    digit = '0' - '9'
    hexnum = ['0x'], { digit | 'a'-'f' | 'A'-'F' } ;
    number = ( digit, { digit } ) | hexnum ;

    group_name = name ;
    event_num = hexnum ;
    subsystem = letter, { letter | digit } ;
    event_name = group_name, '_', name ;
    option = name ;
    event_option = "EVENT_OPTION_", name, '=', number ;
    name = letter, { letter | digit | '_' } ;

    event_list = event_group, { event_group } ;
    event_group = event_group_header, event, { event } ;
    event_group_header = "EVENT_", group_name, event_num, subsystem, { '|', subsystem }, eol ;
    event = [ options ], [ default_options ], umask ;
    options = "OPTIONS_", event_name, option, { '|', option }, eol ;
    default_options = "DEFAULT_OPTIONS_", event_name, event_option, { ',', event_option }, eol ;
    umask = "UMASK_", event_name, hexnum, [ hexnum, [ hexnum ] ], eol ;

    """


    def __init__(self):

        Comment = pythonStyleComment.suppress()
        Hexnum = Literal("0x").suppress() + Word(hexnums).setParseAction(tokenMap(int,16))
        Number = (Optional(Literal("0x")) + Word(hexnums)).setParseAction(
                    lambda s, loc, toks:
                        int(toks[-1], 10 if len(toks) == 1 else 16))
        LinePrefix = lambda pref: Literal(pref) + Literal('_').suppress()
        Name = Word(alphanums + '_.')
        EventName = Name.setResultsName("name")

        self.Evt = (
            LinePrefix("EVENT") + EventName
            - Hexnum
                .setResultsName("number")
                .setName("event code (hex)")
            - delimitedList(Word(alphanums + '_')
                .setName("pipe (|) separated list of subsystem names (PMC, PBOX0|PBOX1 etc.)"),
                delim='|').setResultsName("subsystem")
            ).setParseAction(self._set_event)
        self.Evt.ignore(Comment)

        self.Options = (
            LinePrefix("OPTIONS") + EventName
            - delimitedList(
                        Literal("EVENT_OPTION").suppress() +
                        Literal('_').suppress() +
                        Name.setName("option name"),
                    delim='|')
                .setResultsName("options")
                .setName("pipe (|) separated list of option names")
            - LineEnd()
            ).setParseAction(self._set_options)
        self.Options.ignore(Comment)

        self.DefaultOptions = (
            LinePrefix("DEFAULT_OPTIONS") + EventName
            - delimitedList(
                        Group(
                              Literal("EVENT_OPTION_").suppress() +
                              Name.setName("option name") +
                              Literal('=').suppress() +
                              Number.setName("option value (integer)")
                              ),
                    delim=',')
                .setResultsName("options")
                .setName("option list (comma separated list of NAME=VALUE pairs)")
            - LineEnd()
            ).setParseAction(self._set_default_options)
        self.DefaultOptions.ignore(Comment)

        self.Umask = (
            LinePrefix("UMASK") + EventName
            - OneOrMore(Hexnum)
                .setResultsName("umask")
                .setName("UMASK value (integer)")
            - LineEnd()
            ).setParseAction(self._parse_umask)
        self.Umask.ignore(Comment)

        self.Event = ( self.Umask | self.Evt | self.Options | self.DefaultOptions | Comment ) + StringEnd()

        self.event_head, self.event_count, self.events = None, 0, OrderedDict()
        self.options, self.default_options = None, None


    def _set_event(self, s, loc, token):

        event_head = token.asDict()
        for key in ('number', 'subsystem'):
            value = event_head[key]
            event_head[key] = tuple(value) if len(value) > 1 else value[0]
        event_head.update(line_num = self.line_num)

        if self.event_head:
            if event_head['name'] == self.event_head['name'] and event_head['subsystem'] == self.event_head['subsystem']:
                self.nerrors += 1
                err("\n{}:{}:\nDuplicate event name: {}"
                    "\nPrevious declaration is here:\n{}:{}:\n",
                    self.fname, self.line_num, event_head['name'], self.fname, self.event_head['line_num'])
            if self.event_count == 0:
                if event_head['name'].startswith(self.event_head['name']):
                    self.nerrors += 1
                    err('\n{}:{}:', self.fname, self.line_num)
                    err("Expected: 'UMASK_{}'", event_head['name'])
                    err("Found:    'EVENT_{}'", event_head['name'])
                else:
                    self.nerrors += 1
                    err('\n{}:{}:', self.fname, self.line_num)
                    err("Expected: 'UMASK_{}_....'", self.event_head['name'])
                    err("Found:    'EVENT_{}'", event_head['name'])

        self.event_head, self.event_count = event_head, 0


    def _set_options(self, s, loc, token):

        if self.options is not None:
            self.nerrors += 1
            err("\n{}:{}:\nUnused OPTIONS: {}", self.fname, self.line_num, token['name'])
        if self.opts.naming:
            if not token['name'].startswith(self.event_head['name']):
                self.nerrors += 1
                err("\n{}:{}:\nExpected OPTIONS name starting with: {}", self.fname, self.line_num, self.event_head['name'])
        self.options = token.asDict()['options']


    def _set_default_options(self, s, loc, token):

        if self.default_options is not None:
            self.nerrors += 1
            err("\n{}:{}:\nUnused DEFAULT_OPTIONS: {}", self.fname, self.line_num, token['name'])
        if self.opts.naming:
            if not token['name'].startswith(self.event_head['name']):
                self.nerrors += 1
                err("\n{}:{}:\nExpected DEFAULT_OPTIONS name starting with: {}", self.fname, self.line_num, self.event_head['name'])
        self.default_options = dict(token.asDict()['options'])


    def _parse_umask(self, s, loc, token):

        token = token.asDict()
        if self.opts.naming:
            # TODO: check for naming consistency
            if not token['name'].startswith(self.event_head['name']):
                self.nerrors += 1
                err("\n{}:{}:\nExpected UMASK name starting with: {}", self.fname, self.line_num, self.event_head['name'])

        event = self.event_head.copy()
        event.update(name=token['name'],
                     umask = token['umask'] if len(token['umask']) > 1 else token['umask'][0],
                     line_num = self.line_num
                     )

        key = event['subsystem'], event['name']
        if key in self.events:
            self.nerrors += 1
            err("\n{}:{}:\nDuplicate event name: {}"
                "\nPrevious declaration is here:\n{}:{}:\n",
                self.fname, self.line_num, event['name'], self.fname, self.events[key]['line_num'])

        if self.options is not None:
            event['options'], self.options = self.options, None
        if self.default_options is not None:
            event['default_options'], self.default_options = self.default_options, None

        self.events[key] = event
        self.event_count += 1


    def reset(self, fname, opts):

        self.fname = fname
        self.event_head, self.event_count, self.events = None, 0, OrderedDict()
        self.options, self.default_options = None, None
        self.opts = opts
        self.nerrors = 0
        self.line_num = 0


    def load(self, afile, opts):
        import copy
        lopts = copy.deepcopy(opts)
        lopts.naming = False
        self.check(afile, lopts)
        return self.events


    def check(self, afile, opts):

        self.reset(afile.name, opts)
        try:
            self.parse_lines(afile)
        except Exception as ex:
            self.nerrors += 1
            err("Internal Error: {}", ex)
        return self.nerrors


    def parse_lines(self, afile):

        for l in afile:
            self.line_num += 1
            l = l.strip()
            if not l: continue
            try:
                self.Event.parseString(l)
            except (ParseException, ParseSyntaxException) as pex:
                self.nerrors += 1
                err('\n{}:{}:{}:', afile.name, self.line_num, pex.col)
                err('{}:', pex.msg)
                err('{}', pex.line)
                err("{}^", '-'*pex.col)
        return self.nerrors


    def to_json(self, f, indent=4, sort_keys=True):

        json.dump([ dict([ (k, v) for k, v in V.items() if k != 'line_num' ])
                    for V in self.events.values() ],
                  f, indent=indent, sort_keys=sort_keys)



def test_EventParser(opts):

    import unittest

    class TestEventParser(unittest.TestCase):

        def setUp(self):
            self.parser = EventParser()
            self.reset_parser("", naming=False)

        def reset_parser(self, name, **kw):
            class Opts:
                def __init__(self, **kw):
                    for k, v in kw.items():
                        setattr(self, k, v)
            opts = Opts(**kw)
            self.parser.reset(name, opts)

        def test_event(self):
            res = self.parser.Evt.parseString("EVENT_C_LO_AD_CREDITS_EMPTY             0x22 RBOX0C0|RBOX0C1|RBOX1C0|RBOX1C1").asDict()
            self.assertEqual(res["name"], "C_LO_AD_CREDITS_EMPTY")
            self.assertEqual(res["number"], [0x22])
            self.assertEqual(res["subsystem"], ['RBOX0C0', 'RBOX0C1', 'RBOX1C0', 'RBOX1C1'])


        def test_bad_event(self):
            with self.assertRaises(ParseSyntaxException):
                self.parser.Evt.parseString("EVENT_C_LO_AD_CREDITS_EMPTY 0x22")
            with self.assertRaises(ParseSyntaxException):
                self.parser.Evt.parseString("EVENT_C_LO_AD_CREDITS_EMPTY PMC")


        def test_options(self):
            res = self.parser.Options.parseString("OPTIONS_OFFCORE_RESPONSE_0_OPTIONS                  EVENT_OPTION_MATCH0_MASK|EVENT_OPTION_MATCH1_MASK").asDict()
            self.assertEqual(res["name"], "OFFCORE_RESPONSE_0_OPTIONS")
            self.assertEqual(res["options"], ['MATCH0_MASK', 'MATCH1_MASK'])


        def test_bad_options(self):
            with self.assertRaises(ParseSyntaxException):
                self.parser.Options.parseString("OPTIONS_OFFCORE")
            with self.assertRaises(ParseSyntaxException):
                self.parser.Options.parseString("OPTIONS_OFFCORE EVENT_OPTION_MATCH0_MASK,EVENT_OPTION_MATCH1_MASK")


        def test_default_options(self):
            res = self.parser.DefaultOptions.parseString("DEFAULT_OPTIONS_UOPS_ISSUED_CORE_TOTAL_CYCLES EVENT_OPTION_THRESHOLD=0xA,EVENT_OPTION_INVERT=1,EVENT_OPTION_ANYTHREAD=1").asDict()
            self.assertEqual(res["name"], "UOPS_ISSUED_CORE_TOTAL_CYCLES")
            self.assertEqual(res["options"], [['THRESHOLD', 10], ['INVERT', 1], ['ANYTHREAD', 1]])


        def test_bad_default_options(self):
            with self.assertRaises(ParseSyntaxException):
                self.parser.DefaultOptions.parseString("DEFAULT_OPTIONS_OFFCORE")
            with self.assertRaises(ParseSyntaxException):
                self.parser.DefaultOptions.parseString("DEFAULT_OPTIONS_OFFCORE EVENT_OPTION_MATCH0_MASK|EVENT_OPTION_MATCH1_MASK")


        def test_umask(self):
            self.parser.Umask.setParseAction(lambda s, loc, token: None)
            res = self.parser.Umask.parseString("UMASK_C_LO_AD_CREDITS_EMPTY             0x22 0x33").asDict()
            self.assertEqual(res["name"], "C_LO_AD_CREDITS_EMPTY")
            self.assertEqual(res["umask"], [0x22, 0x33])


        def test_umask_name(self):
            self.reset_parser("test_umask_name", naming=True)
            res = self.parser.parse_lines([
                "EVENT_SIMD_SAT_UOPS_EXEC   0xB1  PMC",
                "UMASK_SIMD_SAT_UOP_EXEC_AR 0x80"])
            self.assertEqual(res, 1)


        def test_options_name(self):
            self.reset_parser("test_options_name", naming=True)
            res = self.parser.parse_lines([
                "EVENT_SIMD_SAT_UOPS_EXEC   0xB1  PMC",
                "OPTIONS_SIMD_SAT_UOP_EXEC_AR EVENT_OPTION_OPCODE_MASK"])
            self.assertEqual(res, 1)


        def test_default_options_name(self):
            self.reset_parser("test_options_name", naming=True)
            res = self.parser.parse_lines([
                "EVENT_SIMD_SAT_UOPS_EXEC   0xB1  PMC",
                "DEFAULT_OPTIONS_SIMD_SAT_UOP_EXEC_AR EVENT_OPTION_DONAME=3"])
            self.assertEqual(res, 1)


    res = unittest.TextTestRunner().run(unittest.TestSuite((
                unittest.makeSuite(TestEventParser),
            )))
    return len(res.errors) + len(res.failures)



def resolve_events_files(opts):

    if not opts.files:
        files = list(opts.input_dir.glob("perfmon_*_events.txt"))
        if not files:
            raise Exception("Couldn't find any events files in '{}', "
                            "please use --input-dir option.\n"
                            "Use `{} events --help` for details."
                            .format(opts.input_dir, sys.argv[0]))
    else:
        files = []
        for f in opts.files:
            f = Path(f).expanduser()
            if f.is_file():
                files.append(f.resolve())
            elif (opts.input_dir / f).is_file():
                files.append((opts.input_dir / f).resolve())
            elif (opts.input_dir / "perfmon_{}_events.txt".format(f)).is_file():
                files.append((opts.input_dir / "perfmon_{}_events.txt".format(f)).resolve())
            else:
                raise Exception("Can't find data file '{}'".format(f))
    return files



def check_events(opts):

    start_logging(args)
    files = resolve_events_files(opts)
    if opts.json:
        opts.output_dir.mkdir(parents=True, exist_ok=True)
    ep = EventParser()
    nerrors = 0
    for fn in files:
        log.info("checking: {}".format(fn))
        with fn.open() as f:
            nerrors += ep.check(f, opts)
        if opts.json:
            o = opts.output_dir / fn.name.replace(fn.suffix, '.json')
            with o.open('w') as jf:
                ep.to_json(jf, indent=4)
    if nerrors:
        log.info("\nFound {} errors".format(nerrors))
    return nerrors > 0



UNITS = [
    's',
    'MHz',
    'MFLOP/s',
    'MUOPS/s',
    'MBytes/s',
    'GBytes',
    '%',
    'J',
    'W',
    'C',
    ]

def miss_brackets_in_units(s):

    for dim in UNITS:
        if ' {} '.format(dim) in s:
            return dim
        if s.endswith(' {}'.format(dim)):
            return dim
    return None



def extract_units(s):

    l = s.split()
    if len(l) > 1 and l[-1].startswith('[') and l[-1].endswith(']'):
        return ' '.join(l[:-1]), l[-1][1:-1]
    else:
        return s, None



class GroupParser():
    """
    eol = '\n' ;
    letter = 'A' - 'Z' | 'a' - 'z' ;
    digit = '0' - '9'
    name = letter, { letter | digit | '_' } ;
    word = -eol, { -eol } ;
    line = word, { word }, eol ;
    op = '*' | '/' | '+' | '-' ;

    short = "SHORT", short_title, eol ;
    noht = "REQUIRE_NOHT", eol ;
    counter_name = name ;
    event_name = name ;
    event = counter_name, event_name, eol ;
    eventset = "EVENTSET", eol, event, { event }, eol ;
    metric_name = letter, { letter | digit | '_' | '(' | ')' | ' ' } ;
    units = UNITS
    metric = metric_name, [ units ], counter_name, {  op, counter_name }, eol ;
    metrics = "METRICS", eol, metric, { metric }, eol ;
    formula = metric_name, [ units ], '=', event_name, {  op,  event_name}, eol ;
    formulas = "Formulas:", eol, formula, { formula } ;
    sep = '--' | '-', eol ;
    long = line, { line } ;
    group = short, [ noht ], eventset, metrics, "LONG", eol, formulas, sep, long
    """

    def __init__(self):

        EOL = LineEnd().suppress()
        Line = LineStart() + SkipTo(LineEnd(), failOn=LineStart()+LineEnd()) + EOL

        short = Keyword('SHORT').suppress() + SkipTo(LineEnd(), failOn=LineStart()+LineEnd()) + EOL
        short.setParseAction(
            lambda s, loc, toks:
                self.group.update(short=toks[0])
        )
        self.parser = short + OneOrMore(EOL)

        noht = Keyword('REQUIRE_NOHT')
        noht.setParseAction(
            lambda s, loc, toks:
                self.group.update(require_noht=True)
        )
        self.parser += Optional(noht() + OneOrMore(EOL))

        eventset = Keyword('EVENTSET').suppress() + EOL + \
        Group(OneOrMore(
            LineStart() + Group(Word(alphanums) + SkipTo(LineEnd(), failOn=LineStart()+LineEnd())
                            ).setParseAction(self.add_event) + EOL
        ))
        self.parser += eventset + OneOrMore(EOL)

        metrics = Keyword('METRICS').suppress() + EOL + \
        Group(OneOrMore(
            LineStart() + Group(OneOrMore(Word(alphanums+'.[]()-+*/')) + SkipTo(LineEnd(), failOn=LineStart()+LineEnd())
                            ).setParseAction(self.add_metric) + EOL
        ))
        self.parser += metrics + OneOrMore(EOL)

        long = Keyword('LONG').suppress() + EOL
        self.parser += long

        Formula = LineStart() + SkipTo(LineEnd(), failOn=LineStart()+(Keyword('-') ^ Keyword('--'))+LineEnd()) + EOL
        formulae = Keyword('Formulas:') + EOL + OneOrMore(Formula().setParseAction(self.add_formula))
        self.parser += formulae

        descr = (Keyword('-') ^ Keyword('--')).suppress() + EOL + Group(OneOrMore(Line()))
        descr.setParseAction(
            lambda s, loc, toks:
                self.group.update(long=' '.join(toks[0]))
        )
        self.parser += descr


    def add_event(self, s, loc, toks):
        log.debug("add_event: |{}|".format(toks[0]))
        self.group['events'].append(dict(counter=toks[0][0], event=toks[0][1]))


    def add_metric(self, s, loc, toks):
        log.debug("add_metric: |{}|".format(toks[0]))
        metric = ' '.join(toks[0][0:-2])
        dim = miss_brackets_in_units(metric)
        if dim:
            raise ParseSyntaxException(s, loc, "expected '[{}]', found '{}'".format(dim, dim))
        metric, units = extract_units(metric)
        formula = toks[0][-2]
        self.group['metrics'].append(dict(metric=metric, formula=formula, units=units))


    def add_formula(self, s, loc, toks):
        if not '=' in toks[0]:
            i = [toks[0].find(x['event']) for x in self.group['events'] if toks[0].find(x['event'])>=0]
            pos = min(i) if i else 0
            while pos > 0 and toks[0][pos] != ' ':
                pos -= 1
            raise ParseSyntaxException(s,loc + pos, "expected '=' sign")
        metric, formula = [ x.strip() for x in toks[0].split('=', 1) ]
        dim = miss_brackets_in_units(metric)
        if dim:
            raise ParseSyntaxException(s, loc, "expected '[{}]', found '{}'".format(dim, dim))
        metric, units = extract_units(metric)
        log.debug("add_formula: |{}| = |{}|".format(metric, formula))
        #if metric in self.group['formulae']:
            #self.nerrors += 1
            #err("\n{}:{}:\nDuplicate metric formula: {}", self.fname, self.line_num, token['name'])
        self.group['formulae'].append(dict(metric=metric, formula=formula, units=units))


    def reset(self, fname, opts):
        self.fname = fname
        self.opts = opts
        self.nerrors = 0
        self.group = dict(
                          short=None,
                          events=[],
                          metrics=[],
                          formulae=[],
                          long=[]
                          )
        self.nerrors = 0


    def check(self, afile, opts):
        self.reset(afile.name, opts)
        try:
            res =self.parser.parseFile(afile)
            log.debug(pformat(self.group))
            log.debug('\ngroup=\n{}'.format(json.dumps(self.group, indent=2)))
        except Exception as pex:
            self.nerrors += 1
            err('\n{}:{}:{}:', self.fname, pex.lineno, pex.col)
            err('{}:', pex.msg)
            err('{}', pex.line)
            err("{}^", '-'*(pex.col-1))
        return self.nerrors


    def to_json(self, f, indent=4, sort_keys=True):

        json.dump(self.group, f, indent=indent, sort_keys=sort_keys)



def resolve_group_files(opts):

    if not opts.files:
        files = list(opts.input_dir.glob("**/*.txt"))
        if not files:
            raise Exception("Couldn't find any group files in '{}', "
                            "please use --input-dir option.\n"
                            "Use `{} groups --help` for details."
                            .format(opts.input_dir, sys.argv[0]))
    else:
        files = []
        for f in opts.files:
            f = Path(f).expanduser()
            _files = []
            if f.is_file():
                _files.append(f.resolve())
            elif (opts.input_dir / f).is_file():
                _files.append((opts.input_dir / f).resolve())
            elif (opts.input_dir / "{}.txt".format(f)).is_file():
                _files.append((opts.input_dir / "{}.txt".format(f)).resolve())
            elif (opts.input_dir / f).is_dir():
                _files = [x.resolve() for x in (opts.input_dir / f).glob("*.txt")]
            if not _files:
                trials = [f, (opts.input_dir / f), opts.input_dir / "{}.txt".format(f), (opts.input_dir / f / '*.txt')]
                raise Exception("Can't find data file:\n\t   '{}'".format("'\n\tor '".join([str(f) for f in trials])))
            files.extend(_files)
    return files



def check_groups(opts):

    start_logging(args)
    files = resolve_group_files(opts)
    gp = GroupParser()
    nerrors = 0
    for fn in files:
        log.info("checking: {}".format(fn))
        with fn.open() as f:
            nerrors += gp.check(f, opts)
        if opts.json:
            o = opts.output_dir / fn.relative_to(opts.input_dir).with_suffix('.json')
            o.parent.mkdir(parents=True, exist_ok=True)
            with o.open('w') as jf:
                gp.to_json(jf, indent=4)
    if nerrors:
        log.info("\nFound {} errors".format(nerrors))
    return nerrors > 0



def start_logging(args):

    log_levels = log.ERROR, log.INFO, log.DEBUG,
    log_level = log_levels[min(args.verbose, len(log_levels) - 1)]
    log.basicConfig(
                    #filename=str(Path(__file__).with_suffix('.log')),
                    #filemode='w',
                    format='%(message)s',
                    level=log_level)



def abs_path(path):
    return Path(path).expanduser().resolve()

if __name__ == "__main__":

    import argparse

    ap = argparse.ArgumentParser(description='Check data files.')
    ap.set_defaults(func=lambda _: ap.print_help())

    subparsers = ap.add_subparsers(title='subcommands',
                                        description="use '<subcmd> --help' for "
                                                    "help on subcommands",
                                        )

    etest = subparsers.add_parser('events', help="test events data files")
    etest.set_defaults(func=check_events)
    etest.add_argument('--verbose', '-v', action='store_true', default=False,
                       help="more verbose output"
                       )
    default = abs_path(__file__).parent / "../src/includes"
    default = default.resolve()
    etest.add_argument('--input-dir', '-d', type=abs_path, default=default,
                help="path to the directory with event data files, "
                     "default: {}".format(default)
                )
    etest.add_argument('files', metavar='FILE', type=str, nargs='*',
                help="event data file to check; "
                     "will be looked relative to current directory first, "
                     "then relative to the INPUT_DIR and then "
                     "INPUT_DIR/perfmon_{FILE}_events.txt filre will be tried."
                )
    etest.add_argument('--json', '-j', action='store_true', default=False,
                help="dump event descriptions in JSON format"
                )
    etest.add_argument('--output-dir', '-o', type=abs_path, default=abs_path('.'),
                help="directory where to put output JSON files, "
                     "default: current directory"
                )
    etest.add_argument('--naming', '-n', action='store_true', default=False,
                help="check for naming consistency, i.e. name in the "
                     "UMASK, OPTIONS and DEFAULT_OPTIONS lines should "
                     "have the name from the EVENT line as a prefix."
                )

    self_test = subparsers.add_parser('self', help="self test for the events checker")
    self_test.set_defaults(func=test_EventParser)

    gtest = subparsers.add_parser('groups', help="test group ddescriptions")
    gtest.set_defaults(func=check_groups)
    gtest.add_argument('--verbose', '-v', action='count', default=0)
    default = abs_path(__file__).parent / "../groups"
    if default.exists():
        default = default.resolve()
    else:
        default = abs_path(__file__).parent
    gtest.add_argument('--input-dir', '-d', type=abs_path,
                        default=default,
                        help='path to the directory with groups data files, default: {}'.format(default))
    gtest.add_argument('files', metavar='FILE', type=str, nargs='*',
                    help='group data file to check')
    gtest.add_argument('--json', '-j', action='store_true', default=False,
                    help="dump group descriptions in JSON format")
    gtest.add_argument('--output-dir', '-o', type=abs_path,
                        default=abs_path('.'),
                        help='directory where to put output JSON files, default: current directory')

    args = ap.parse_args()

    try:
        rc = args.func(args)
    except Exception as ex:
        err("{}", ex)
        sys.exit(1)
    sys.exit(rc)



#!/usr/bin/env python

# =======================================================================================
#
#      Filename:  check_group_files.py
#
#      Description:  Basic checks for performance group files
#
#      Author:   Thomas Gruber (tr), thomas.roehl@googlemail.com
#      Project:  likwid
#
#      Copyright (C) 2016 RRZE, University Erlangen-Nuremberg
#
#      This program is free software: you can redistribute it and/or modify it under
#      the terms of the GNU General Public License as published by the Free Software
#      Foundation, either version 3 of the License, or (at your option) any later
#      version.
#
#      This program is distributed in the hope that it will be useful, but WITHOUT ANY
#      WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
#      PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License along with
#      this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =======================================================================================

import sys, os, re, glob, os.path, subprocess

SRCPATH="/home/rrze/unrz/unrz139/Work/likwid"
GROUPPATH=os.path.join(SRCPATH, "groups")
EVENTHEADERPATH=os.path.join(SRCPATH, "src/includes")
PERFCTR="likwid-perfctr"

def get_valid_shorts():
    l = os.listdir(GROUPPATH)
    return l


def get_local_arch():
    p = subprocess.Popen("%s -i" % PERFCTR, stdout=subprocess.PIPE, shell=True)
    stdoutdata, stderrdata = p.communicate()
    if p.returncode == 0:
        for l in stdoutdata.split("\n"):
            if l.startswith("CPU short:"):
                return re.split("\s+", l)[-1]


def get_all_events_local_arch():
    events = {}
    p = subprocess.Popen("%s -e" % PERFCTR, stdout=subprocess.PIPE, shell=True)
    stdoutdata, stderrdata = p.communicate()
    if p.returncode == 0:
        parse = False

        for l in stdoutdata.split("\n"):
            if l.startswith("Event tags"):
                parse = True
                continue
            if parse:
                if not l.strip(): continue
                llist = re.split(",\s+", l)
                events[llist[0]] = llist[-1]
    return events

def get_all_events_given_arch(arch):
    events = {}
    efile = os.path.join(EVENTHEADERPATH, "perfmon_%s_events.txt" % arch)
    if os.path.exists(efile):
        f = open(efile, "r")
        raw = f.read().strip().split("\n")
        f.close()
        limit = None
        for l in raw:
            if l.startswith("EVENT_"):
                limit = re.split("\s+", l)[-1]
            if l.startswith("UMASK_"):
                e = re.match("UMASK_([\w\d_]+)", l)
                if e:
                    events[e.group(1)] = limit
    return events

def get_all_counters_local_arch():
    counters = []
    p = subprocess.Popen("%s -e" % PERFCTR, stdout=subprocess.PIPE, shell=True)
    stdoutdata, stderrdata = p.communicate()
    if p.returncode == 0:
        parse = False

        for l in stdoutdata.split("\n"):
            if l.startswith("Counter tags"):
                parse = True
                continue
            if parse:
                if not l.strip(): continue
                llist = re.split(",\s+", l)
                counters.append(llist[0])
    return counters

def get_all_counters_given_arch(arch):
    counters = []
    cfile = os.path.join(EVENTHEADERPATH, "perfmon_%s_counters.h" % arch)
    if os.path.exists(cfile):
        f = open(cfile, "r")
        raw = f.read().strip().split("\n")
        f.close()
        parse = False
        for l in raw:
            if l.startswith("static RegisterMap"):
                parse = True
                continue
            if l.startswith("};"):
                parse = False
            if parse:
                c = re.match("^\s+{\"([\w\d]+)\"", l)
                if c:
                    counters.append(c.group(1))
    return counters

def get_all_groupfiles(arch):
    path=os.path.join(GROUPPATH, arch)
    if not os.path.exists(path):
        print("Cannot find group path %s" % path)
        return []
    grouplist = glob.glob(path+"/*")
    return grouplist

def check_short(gfile):
    if not os.path.exists(gfile):
        print("Cannot find group file %s" % gfile)
        return False
    with open(gfile) as f:
        inf = f.read().split("\n")
        for l in inf:
            if l.startswith("SHORT"):
                if len(re.split("\s+", l)) > 1:
                    return True
                else:
                    return False
    return False

def check_eventset(gfile, allevents, allcounters):
    events = {}
    noht = False
    if not os.path.exists(gfile):
        print("Cannot find group file %s" % gfile)
        return False
    with open(gfile) as f:
        parse = False
        inf = f.read().split("\n")
        for l in inf:
            if l.startswith("REQUIRE_NOHT"):
                noht = True
            if l.startswith("EVENTSET"):
                parse = True
                continue
            if l.startswith("METRICS"):
                parse = False
                continue
            if parse:
                if not l.strip(): continue
                elist = re.split("\s+", l)
                o = None
                c = elist[0]
                e = elist[1]
                if ":" in c:
                    tmp = c.split(":")
                    c = tmp[0]
                    o = ":".join(tmp[1:])
                if not events.has_key(c):
                    events[c] = e
                else:
                    print("Counter register used twice: %s and %s" % (e, events[c]))
                    return False
    for c in events.keys():
        if c not in allcounters and not noht:
            print("Counter register %s does not exist" % c)
            return False
        elif c not in allcounters:
            print("Group requires HyperThreading to be off!")
            return True
        if events[c] not in allevents.keys():
            print("Event %s unknown" % events[c])
            return False
    return True

def check_metrics(gfile):
    if not os.path.exists(gfile):
        print("Cannot find group file %s" % gfile)
        return False
    metrics = {}
    with open(gfile) as f:
        parse = False
        inf = f.read().split("\n")
        for l in inf:
            if l.startswith("METRICS"):
                parse = True
                continue
            if l.startswith("LONG"):
                parse = False
                continue
            if parse and len(l) > 0:
                llist = re.split("\s+", l)
                name = " ".join(llist[:-1])
                if "[G" in l and not re.match("1[\.\d]*E[-+][\d]*9", llist[-1]):
                    print("Wrong unit? %s" % l)
                if "[M" in l and not "[MHz]" in l and not re.match("1[\.\d]*E[-+][\d]*6", llist[-1]):
                    print("Wrong unit? %s" % l)
                if "[%]" in l and not ("100*" in l or "*100" in l):
                    print("Scaling factor missing? %s" % l)
                metrics[" ".join(llist[:-1])] = llist[-1]
    if len(metrics.keys()) > 0:
        return True
    return False

def check_long(gfile):
    if not os.path.exists(gfile):
        print("Cannot find group file %s" % gfile)
        return False
    longlines = []
    with open(gfile) as f:
        parse = False
        inf = f.read().split("\n")
        for l in inf:
            if l.startswith("LONG"):
                parse = True
                continue
            if parse and len(l) > 0:
                longlines.append(l)
    if len(longlines) > 0:
        return True
    return False

arch = None
if len(sys.argv) == 2:
    arch = sys.argv[1]
    if arch == "-h" or arch == "--help":
        print("Checks performance group files for some basic problems:")
        print("\t- Is a short description defined?")
        print("\t- Is an eventset defined?")
        print("\t- Are all counters in the eventset available?")
        print("\t- Are all events in the eventset available?")
        print("\t- Is the scaling factor correct accoring to the unit in metric name")
        print("\t- Is there a long description?")
        print("")
        print("If no command line argument is given, all architectures are tested.")
        print("")
        print("If the first argument is a valid architecture string, the lists are")
        print("filled from definitions from source.")
        print("")
        print("If first argument is 'local' the current system is checked. This can")
        print("throw more errors as the other modes as the output of likwid-perfctr -e")
        print("contains only counters and events that are accessible on the current system.")
        print("")
        print("Available architectures:")
        print(", ".join(get_valid_shorts()))
        sys.exit(0)
    if arch != "local" and arch not in get_valid_shorts():
        print("Given arch not available")
        sys.exit(1)

alist = []
if arch == "local":
    events = get_all_events_local_arch()
    counters = get_all_counters_local_arch()
    alist = [get_local_arch()]
elif arch != None:
    ea = arch
    ca = arch
    if arch == "broadwellD":
        ea = "broadwelld"
        ca = "broadwelld"
    if arch == "pentiumm":
        ea = "pm"
        ca = "pm"
    if arch == "westmere":
        ca = "nehalem"
    if arch == "atom":
        ca = "core2"
    if arch == "k8":
        ca = "k10"
    events = get_all_events_given_arch(ea)
    counters = get_all_counters_given_arch(ca)
    alist = [arch]
else:
    alist = get_valid_shorts()

for a in alist:
    print("Checking architecture %s" % a)
    glist = get_all_groupfiles(a)
    if len(alist) > 1:
        ea = a
        ca = a
        if a == "broadwellD":
            ea = "broadwelld"
            ca = "broadwelld"
        if a == "pentiumm":
            ea = "pm"
            ca = "pm"
        if a == "westmere":
            ca = "nehalem"
        if a == "atom":
            ca = "core2"
        if a == "k8":
            ca = "k10"
        events = get_all_events_given_arch(ea)
        counters = get_all_counters_given_arch(ca)
    for f in glist:
        g = os.path.basename(f).split(".")[0]
        print("Checking group %s" % g)
        if not check_short(f):
            print("Short failure in group file %s" % f)
        if not check_eventset(f, events, counters):
            print("Eventset failure in group file %s" % f)
        if not check_metrics(f):
            print("Metrics failure in group file %s" % f)
        if not check_long(f):
            print("Long failure in group file %s" % f)
    print("")



#!/bin/env python3
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple script for generating the extract.h header's boilerplate.

Update it by running
$ cwisstable/internal/extract.py
"""

import os
from pathlib import Path

DEPTH = 64 
KEYS = [
  'obj_copy', 'obj_dtor',
  'key_hash', 'key_eq',
  'alloc_alloc', 'alloc_free',
  
  'slot_size', 'slot_align', 'slot_init',
  'slot_transfer', 'slot_get', 'slot_dtor',
  'modifiers',
]
FILE = Path(__file__).parent / 'extract.h'

def main():
  text = FILE.read_text()
  end = text.find('// !!!')

  lines = []
  lines.append(text[:end] + '// !!!')
  lines.append('')

  for key in KEYS:
    lines.append(f'#define CWISS_EXTRACT_{key}(key_, val_) CWISS_EXTRACT_{key}Z##key_')
    lines.append(f'#define CWISS_EXTRACT_{key}Z{key} CWISS_NOTHING, CWISS_NOTHING, CWISS_NOTHING')
  lines.append('')

  for i in range(0, DEPTH):
    lines.append(f'#define CWISS_EXTRACT{i:02X}(needle_, kv_, ...) CWISS_SELECT{i:02X}(needle_ kv_, CWISS_EXTRACT_VALUE, kv_, CWISS_EXTRACT{i+1:02X}, (needle_, __VA_ARGS__), CWISS_NOTHING)')
  lines.append('')
  for i in range(0, DEPTH):
    lines.append(f'#define CWISS_SELECT{i:02X}(x_, ...) CWISS_SELECT{i:02X}_(x_, __VA_ARGS__)')
  lines.append('')
  for i in range(0, DEPTH):
    lines.append(f'#define CWISS_SELECT{i:02X}_(ignored_, _call_, _args_, call_, args_, ...) call_ args_')
  lines.append('')
  lines.append('#endif  // CWISSTABLE_INTERNAL_EXTRACT_H_')

  FILE.write_text('\n'.join(lines) + '\n')

if __name__ == '__main__': main()


#!/usr/bin/env python

import os, sys, os.path
import re
import subprocess
import socket
import stat
import getopt

# Needed for Wiki page output
import glob
import statistics

bench_plain = "./likwid-bench-plain"
bench_marker = "/home/hpc/unrz/unrz139/Apps-arm/bin/likwid-bench"
bench_papi = "./likwid-bench-papi"
perfctr = "/home/hpc/unrz/unrz139/Apps-arm/bin/likwid-perfctr"
topology = "/home/hpc/unrz/unrz139/Apps-arm/bin/likwid-topology"
topology_name = re.compile("^CPU name:\s+(.*)")
topology_stepping = re.compile("^CPU stepping:\s+(\d*)")
topology_type = re.compile("^CPU type:\s+(.*)")
topology_sockets = re.compile("^Sockets:\s+(\d+)")
topology_corespersocket = re.compile("^Cores per socket:\s+(\d+)")
topology_threadspercore = re.compile("^Threads per core:\s+(\d+)")
testlist = "SET.txt"
testfolder = "TESTS"
resultfolder = "RESULTS"
hostname = socket.gethostname()
picture_base = ".."
topology_outputfile = "topology.dat"
nrThreads = 1

gnu_colors = ["red","blue","green","black"]#,"brown", "gray","violet", "cyan", "magenta","orange","#4B0082","#800000","turquoise","#006400","yellow"]
gnu_marks = [5,13,9,2]#,3,4,6,7,8,9,10,11,12,14,15]

units = { "L2" : "MByte/s", "L3" : "MByte/s", "MEM" : "MByte/s", "HA" : "MByte/s",
          "FLOPS_SP" : "MFLOP/s", "FLOPS_DP" : "MFLOP/s", "FLOPS_AVX" : "MFLOP/s",
          "DATA": "Load/Store ratio", "BRANCH" : "Instructions per branch",
          "CLOCK" : "Instructions", "UOPS" : "UOPs"}
translate_group = {"CLOCK" : "INST_RETIRED_ANY", "UOPS" : "UOPS_RETIRED_ANY"}

wiki = False
papi = False
only_wiki = False
sets = []
out_pgf = False
out_gnuplot = False
out_grace = False
scriptfilename = "create_plots.sh"
out_script = False
test_set = {}
plain_set = {}
corrected_set = {}
marker_set = {}
papi_set = {}

if not os.path.exists(bench_marker):
    print "Please run make before using likwid-accuracy.py"
    sys.exit(1)
if not os.path.exists(perfctr):
    print "Cannot find likwid-perfctr"
    sys.exit(1)


def usage():
    print "Execute and evaluate accuracy tests for LIKWID with likwid-bench and likwid-perfctr"
    print
    print "-h/--help:\tPrint this help text"
    print "-s/--sets:\tSpecifiy testgroups (comma separated). Can also be set in SET.txt"
#    print "--wiki:\t\tBesides testing write out results in Google code wiki syntax"
#    print "--only_wiki:\tDo not run benchmarks, read results from file and write out results in Google code wiki syntax"
    print "-c <nrThreads>:\tSet number of threads. The accuracy tool uses the E notation of likwid like E:N:<nrThreads>:1:2. Default is 1 thread."
    print "Picture options:"
    print "--pgf:\t\tCreate TeX document for each test with PGFPlot"
    print "--gnuplot:\tCreate GNUPlot script for each test"
    print "--grace:\tCreate Grace script that can be evaluated with gracebat"
    print "--script:\tActivate recording of commands in a bash script"
    print "--scriptname:\tRecord commands to create pictures in file (default: %s)" % (os.path.join(destfolder,scriptfilename))


def get_groups():
    groups = {}
    p = subprocess.Popen(perfctr+" -a", shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    p.wait()
    if p.returncode != 0:
        return groups
    for line in p.stdout.read().split("\n"):
        if line.startswith("-") or not line.strip(): continue
        if line.startswith("Available"): continue
        linelist = re.split("\s+", line.strip())
        name = linelist[0]
        description = " ".join(linelist[1:])
        groups[name] = description
    return groups

def get_test_groups(groupdict):
    groups = {}
    if len(sets) > 0:
        setlist = sets
    else:
        setfp = open("SET.txt",'r')
        setlist = setfp.read().strip().split("\n")
        setfp.close()

    filelist = glob.glob(testfolder+"/*.txt")
    for name in setlist:
        if name in get_groups():
            tests = []
            file = os.path.join(testfolder, name) + ".txt"
            if not os.path.exists(file): continue
            fp = open(file,'r')
            finput = fp.read().strip().split("\n")
            fp.close()
            for line in finput:
                if line.startswith("TEST"):
                    tests.append(line.split(" ")[1])
            groups[name] = tests


    return groups

def get_avail_tests():
    tests = []
    p = subprocess.Popen(bench_marker+" -a", shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    p.wait()
    if p.returncode != 0:
        return tests
    for line in p.stdout.read().split("\n"):
        m = re.match("(.*) - (.*)", line)
        if m:
            name, desc = m.groups()
            tests.append(name)
    return tests

def write_topology(path):
    try:
        f = open(os.path.join(path, topology_outputfile),"w")
    except:
        print "Cannot write topology file %s" % (os.path.join(path, topology_outputfile),)
        return
    p = subprocess.Popen(topology, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    p.wait()
    if p.returncode != 0:
        return
    f.write(p.stdout.read())
    f.close()

def approx(in1, in2):
    if in1 > (0.95*in2) or in1 < (1.05*in2):
        return 1
    return 0

def legend(file1, file2):
    input1 = []
    input2 = []
    numbers1 = []
    numbers2 = []
    try:
        f=open(file1,"r")
        input1 = f.read().strip().split("\n")
        f.close()
    except:
        print "Cannot open file "+file1
    try:
        f=open(file2,"r")
        input2 = f.read().strip().split("\n")
        f.close()
    except:
        print "Cannot open file "+file2
    if len(input1) == 0 and len(input2) == 0:
        return "no"
    for line in input1:
        numbers1.append(line.split(" ")[1])
    for line in input2:
        numbers2.append(line.split(" ")[1])
    if float(numbers1[1]) > float(numbers1[-1]) and float(numbers2[1]) > float(numbers2[-1]):
        return "no"
    elif float(numbers1[1]) < float(numbers1[-1]) and float(numbers2[1]) < float(numbers2[-1]):
        return "so"
    elif approx(float(numbers1[1]), float(numbers1[-1])) and approx(float(numbers2[1]), float(numbers2[-1])):
        return "so"
    return "no"


def write_pgf(group, test, plain_file, marker_file, scale=0.0,papi_file=None, execute=False, script=None):
    printgrp = group
    if translate_group.has_key(group):
        printgrp = translate_group[group]
    filename = os.path.join(destfolder,printgrp+"_"+test+".tex")
    sizelist = []
    sizeindex = []
    lentry = "north east"
    if legend(plain_file, marker_file) == "so":
        lentry = "south east"
    for i,variant in enumerate(test_set[group][test]["variants"]):
        sizelist.append(variant)
        sizeindex.append(str((i+0.5)*test_set[group][test]["RUNS"]))
    fp = open(filename,'w')
    fp.write("\documentclass{article}\n")
    fp.write("\usepackage{pgfplots}\n")
    fp.write("\\begin{document}\n")
    fp.write("% cut from here\n")
    fp.write("\\begin{tikzpicture}\n")
    fp.write("\\begin{axis}[xmin=0,xmax=%d,xlabel={Size - %d runs each}, ylabel={%s},title={Group %s - Test %s},legend pos=%s,xtick=data,width=.75\\textwidth,xticklabels={%s},xtick={%s}]\n" % (test_set[group][test]["RUNS"]*len(test_set[group][test]["variants"]),test_set[group][test]["RUNS"],units[group],printgrp.replace("_","\_"),test.replace("_","\_"),lentry,",".join(sizelist),",".join(sizeindex)))
    fp.write("\\addplot+[red,mark=square*,mark options={draw=red, fill=red}] table {%s};\n" % (os.path.basename(plain_file),))
    fp.write("\\addlegendentry{bench};\n")
    if scale > 0.0:
        fp.write("\\addplot+[blue,mark=*,mark options={draw=blue, fill=blue}] table[x index=0, y expr=\\thisrowno{1}*%f] {%s};\n" % (scale, os.path.basename(plain_file),))
        fp.write("\\addlegendentry{scaled bench};\n")
    fp.write("\\addplot+[green,mark=diamond*,mark options={draw=green, fill=green}] table {%s};\n" % (os.path.basename(marker_file),))
    fp.write("\\addlegendentry{perfctr};\n")
    if papi and papi_file:
        fp.write("\\addplot+[black,mark=triangle*,mark options={draw=black, fill=black}] table {%s};\n" % (os.path.basename(papi_file),))
        fp.write("\\addlegendentry{papi};\n")
    fp.write("\\end{axis}\n")
    fp.write("\\end{tikzpicture}\n")
    fp.write("% stop cutting here\n")
    fp.write("\\end{document}\n")
    fp.close()
    if execute:
        cmd = "cd %s && pdflatex %s && cd -" % (os.path.dirname(filename), os.path.basename(filename),)
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.wait()
        if p.returncode != 0:
            print p.stdout.read()
        p.stdout.close()
    if script:
        script.write("pdflatex %s\n" % (os.path.basename(filename),))
    return filename

def write_gnuplot(group, test, plain_file, marker_file, scale = 1.0, papi_file=None, execute=False, script=None):
    printgrp = group
    if translate_group.has_key(group):
        printgrp = translate_group[group]
    filename = os.path.join(destfolder,printgrp+"_"+test+".plot")
    fp = open(filename,'w')
    for i,color in enumerate(gnu_colors):
        fp.write("set style line %d linetype 1 linecolor rgb '%s' lw 2 pt %s\n" % (i+1, color,gnu_marks[i]))
    fp.write("set terminal jpeg enhanced\n")
    fp.write("set encoding utf8\n")
    fp.write("set title 'Group %s - Test %s'\n" % (printgrp.replace("_", "\_"), test.replace("_", "\_"),))
    if legend(plain_file, marker_file) == "no":
        fp.write("set key top right\n")
    else:
        fp.write("set key bottom right\n")
    fp.write("set output '%s'\n" % (os.path.basename(os.path.join(destfolder,printgrp+"_"+test+".jpg")),))
    fp.write("set xlabel 'Size - %d runs each'\n" % (test_set[group][test]["RUNS"],))
    fp.write("set ylabel '%s'\n" % (units[group],))
    fp.write("set yrange  [0:]\n")
    #fp.write("set xtics 0,%d,%d\n" % (test_set[group][test]["RUNS"], test_set[group][test]["RUNS"]*len(test_set[group][test]["variants"]),))
    fp.write("set xtics %d\n" % (test_set[group][test]["RUNS"]*len(test_set[group][test]["variants"]),))
    for i,variant in enumerate(test_set[group][test]["variants"]):
        fp.write("set xtics add (\"%s\" %f)\n" % (variant, (i*test_set[group][test]["RUNS"])+(0.5*test_set[group][test]["RUNS"]),))
    plot_string = "plot '%s' using 1:3 title 'Reference' with linespoints ls 1, \\\n"  % (os.path.basename(plain_file),)
    if scale > 0.0:
        plot_string = plot_string+" '%s' using 1:($3*%f) title 'Scaled Reference' with linespoints ls 2, \\\n" % (os.path.basename(plain_file), scale,)
    plot_string = plot_string+" '%s' using 1:3 title 'Measurement (LIKWID)' with linespoints ls 3" % (os.path.basename(marker_file),)
    if papi and papi_file:
        plot_string += ", \\\n '%s' using 1:3 title 'Measurement (PAPI)' with linespoints ls 4\n" % (os.path.basename(papi_file),)
    fp.write(plot_string+"\n")
    fp.close()
    if execute:
        cmd = "cd %s && gnuplot %s && cd -" % (os.path.dirname(filename), os.path.basename(filename),)
        print cmd
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.wait()
        if p.returncode != 0:
            print p.stdout.read()
        p.stdout.close()
    if script:
        script.write("gnuplot %s\n" % (os.path.basename(filename),))
    return filename

def write_grace(group, test, plain_file, correct_file, marker_file, papi_file=None, execute=False, script=None):
    printgrp = group
    if translate_group.has_key(group):
        printgrp = translate_group[group]
    filename = os.path.join(destfolder,printgrp+"_"+test+".bat")
    agrname = os.path.join(destfolder,printgrp+"_"+test+".agr")
    pngname = os.path.join(destfolder,printgrp+"_"+test+".png")
    if execute or script:
        plain_file = os.path.basename(plain_file)
        marker_file = os.path.basename(marker_file)
        correct_file = os.path.basename(correct_file)
        if papi_file: papi_file = os.path.basename(papi_file)
        pngname = os.path.basename(pngname)
        agrname = os.path.basename(agrname)
    cmd_options = "-autoscale xy -nxy %s -nxy %s -nxy %s " % (plain_file, correct_file, marker_file,)
    if papi and papi_file:
        cmd_options += "-nxy %s " % (papi_file,)
    out_options = "-hdevice PNG -printfile %s " % (pngname,)
    out_options += "-saveall %s" % (agrname,)
    fp = open(filename,'w')
    fp.write("title \"Group %s - Test %s\"\n" % (printgrp, test,))
    fp.write("xaxis label \"Run\"\n")
    fp.write("xaxis label char size 1.2\n")
    fp.write("xaxis ticklabel char size 1.2\n" % (units[group],))
    fp.write("yaxis label \"%s\"\n")
    fp.write("yaxis label char size 1.2\n")
    fp.write("yaxis ticklabel char size 1.2\n")
    if legend(plain_file, marker_file) == "no":
        fp.write("legend 0.8,0.7\n")
    else:
        fp.write("legend 0.2,0.7\n")
    fp.write("s0 legend \"bench\"\n")
    fp.write("s0 symbol 2\n")
    fp.write("s0 symbol size 1\n")
    fp.write("s0 symbol color 2\n")
    fp.write("s0 symbol pattern 1\n")
    fp.write("s0 symbol fill color 2\n")
    fp.write("s0 symbol fill pattern 1\n")
    fp.write("s0 symbol linewidth 2\n")
    fp.write("s0 symbol linestyle 1\n")
    fp.write("s0 line type 1\n")
    fp.write("s0 line color 2\n")
    fp.write("s0 line linestyle 1\n")
    fp.write("s0 line linewidth 2\n")
    fp.write("s0 line pattern 1\n")
    fp.write("s1 legend \"scaled bench\"\n")
    fp.write("s1 symbol 3\n")
    fp.write("s1 symbol size 1\n")
    fp.write("s1 symbol color 4\n")
    fp.write("s1 symbol pattern 1\n")
    fp.write("s1 symbol fill color 4\n")
    fp.write("s1 symbol fill pattern 1\n")
    fp.write("s1 symbol linewidth 2\n")
    fp.write("s1 symbol linestyle 1\n")
    fp.write("s1 line type 1\n")
    fp.write("s1 line color 4\n")
    fp.write("s1 line linestyle 1\n")
    fp.write("s1 line linewidth 2\n")
    fp.write("s1 line pattern 1\n")
    fp.write("s2 legend \"perfctr\"\n")
    fp.write("s2 symbol 4\n")
    fp.write("s2 symbol size 1\n")
    fp.write("s2 symbol color 3\n")
    fp.write("s2 symbol pattern 1\n")
    fp.write("s2 symbol fill color 3\n")
    fp.write("s2 symbol fill pattern 1\n")
    fp.write("s2 symbol linewidth 2\n")
    fp.write("s2 symbol linestyle 1\n")
    fp.write("s2 line type 1\n")
    fp.write("s2 line color 3\n")
    fp.write("s2 line linestyle 1\n")
    fp.write("s2 line linewidth 2\n")
    fp.write("s2 line pattern 1\n")
    if papi and papi_file:
        fp.write("s3 legend \"papi\"\n")
        fp.write("s3 symbol 5\n")
        fp.write("s3 symbol size 1\n")
        fp.write("s3 symbol color \"black\"\n")
        fp.write("s3 symbol pattern 1\n")
        fp.write("s3 symbol fill color \"black\"\n")
        fp.write("s3 symbol fill pattern 1\n")
        fp.write("s3 symbol linewidth 2\n")
        fp.write("s3 symbol linestyle 1\n")
        fp.write("s3 line type 1\n")
        fp.write("s3 line color \"black\"\n")
        fp.write("s3 line linestyle 1\n")
        fp.write("s3 line linewidth 2\n")
        fp.write("s3 line pattern 1\n")
    fp.close()
    if execute:
        cmd = "cd %s && gracebat %s -param %s %s && cd -" % (os.path.dirname(filename), cmd_options, os.path.basename(filename),out_options,)
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        p.wait()
        if p.returncode != 0:
            print p.stdout.read()
        p.stdout.close()
    if script:
        script.write("gracebat %s -param %s %s\n" % (cmd_options, os.path.basename(filename),out_options,))
    return filename

def get_bytes(s):
    m = re.match("(\d+)[bB]", s)
    if m:
        return int(m.group(1))
    m = re.match("(\d+)[kK][bB]", s)
    if m:
        return int(m.group(1))*1024
    m = re.match("(\d+)[mM][bB]", s)
    if m:
        return int(m.group(1))*1024*1024
    m = re.match("(\d+)[gG][bB]", s)
    if m:
        return int(m.group(1))*1024*1024*1024

def get_sizestr(b):
    bint = int(b)
    if int(bint / 1024.0) == 0:
        return "%dB" % (bint)
    if int(bint / (1024.0*1024.0)) == 0:
        if bint % (1024) == 0:
            return "%dkB" % (bint/(1024))
        else:
            return "%dB" % (bint)
    if int(bint / (1024.0*1024.0*1024.0)) == 0:
        if bint % (1024*1024) == 0:
            return "%dMB" % (bint/(1024*1024))
        else:
            return "%dkB" % (bint/(1024))
    if int(bint / (1024.0*1024.0*1024.0*1024.0)) == 0:
        if bint % (1024*1024*1024) == 0:
            return "%dGB" % (bint/(1024*1024*1024))
        else:
            return "%dMB" % (bint/(1024*1024))

def get_variants(start, end, step):
    bstart = get_bytes(start)
    bend = get_bytes(end)
    bstep = get_bytes(step)
    return [ get_sizestr(x) for x in range(bstart, bend, bstep) ] + [end]

try:
    opts, args = getopt.getopt(sys.argv[1:], "hs:c:", ["help", "sets=","script","scriptname=","wiki","only_wiki=","pgf","gnuplot","grace","papi"])
except getopt.GetoptError as err:
    print str(err)
    usage()
    sys.exit(2)

if len(opts) == 0:
    usage()
    sys.exit(1)

for o, a in opts:
    if o in ("-h","--help"):
        usage()
        sys.exit(0)
    if o == "--wiki":
        wiki = True
    if o == "--only_wiki":
        only_wiki = True
        hostname = a
    if o == "--papi":
        papi = True
    if o == "-c":
        try:
            nrThreads = int(a)
        except:
            print "Argument to -c not valid. Must be a number"
            sys.exit(1)
    if o == "--pgf":
        out_pgf = True
    if o == "--gnuplot":
        out_gnuplot = True
    if o == "--grace":
        out_grace = True
    if o in ("-s","--sets"):
        sets = a.split(",")
    if o == "--script":
        out_script = True
    if o == "--scriptname":
        scriptfilename = a

if len(sets) == 0 and not os.path.exists(testlist):
    print "Cannot find file %s containing list of testgroups" % (testlist,)
    sys.exit(1)
if not os.path.exists(testfolder):
    print "Cannot find folder %s containing the testgroups" % (testfolder,)
    sys.exit(1)


if len(sets) == 0:
    fp = open(testlist,'r')
    tmp = fp.read().split("\n")
    for item in tmp:
        if not item.strip() or item.startswith("#"): continue
        sets.append(item)
    fp.close()
for line in sets:
    if not line.strip() or line.startswith("#"): continue
    filename = "%s/%s.txt" % (testfolder,line.strip(),)
    if os.path.exists(filename):
        groupname = line.strip()
        testfp = open(filename,'r')
        for line in testfp.read().split("\n"):
            if line.startswith("GROUP"):
                match = re.match("^GROUP\s+(\.+)", line)
                if match:
                    groupname = match.group(1)
                    break
        testfp.close()
        test_set[groupname] = {}
        plain_set[groupname] = {}
        corrected_set[groupname] = {}
        marker_set[groupname] = {}
        papi_set[groupname] = {}
        testfp = open(filename,'r')
        test = None
        for i,testline in enumerate(testfp.read().split("\n")):
            if test and not testline.strip(): test = None
            if testline.startswith("REGEX_BENCH"):
                test_set[groupname]["REGEX_BENCH"] = re.compile(" ".join(testline.split(" ")[1:]))
            if testline.startswith("REGEX_PERF"):
                test_set[groupname]["REGEX_PERF"] = re.compile(" ".join(testline.split(" ")[1:]))
            if testline.startswith("REGEX_PAPI"):
                test_set[groupname]["REGEX_PAPI"] = re.compile(" ".join(testline.split(" ")[1:]))
            if testline.startswith("TEST"):
                test = testline.split(" ")[1]
                test_set[groupname][test] = {}
                test_set[groupname][test]["WA_FACTOR"] = 0.0
                plain_set[groupname][test] = {}
                corrected_set[groupname][test] = {}
                marker_set[groupname][test] = {}
                papi_set[groupname][test] = {}
            if testline.startswith("RUNS") and test:
                test_set[groupname][test]["RUNS"] = int(testline.split(" ")[1])
            if testline.startswith("WA_FACTOR") and test:
                test_set[groupname][test]["WA_FACTOR"] = float(testline.split(" ")[1])
            if testline.startswith("VARIANT") and test:
                linelist = re.split("\s+",testline);
                variant = linelist[1]
                if not test_set[groupname][test].has_key("variants"):
                    test_set[groupname][test]["variants"] = []
                test_set[groupname][test][variant] = linelist[2]
                test_set[groupname][test]["variants"].append(linelist[1])
                plain_set[groupname][test][variant] = []
                corrected_set[groupname][test][variant] = []
                marker_set[groupname][test][variant] = []
                papi_set[groupname][test][variant] = []
            if testline.startswith("RANGE") and test:
                linelist = re.split("\s+",testline);
                start = linelist[1]
                end = linelist[2]
                step = linelist[3]
                if not test_set[groupname][test].has_key("variants"):
                    test_set[groupname][test]["variants"] = []
                variants = get_variants(start, end, step)
                for variant in variants:
                    test_set[groupname][test][variant] = "1"
                    test_set[groupname][test]["variants"].append(variant)
                    plain_set[groupname][test][variant] = []
                    corrected_set[groupname][test][variant] = []
                    marker_set[groupname][test][variant] = []
                    papi_set[groupname][test][variant] = []
        testfp.close()


if len(test_set.keys()) == 0:
    print "Cannot find any group in %s" % (testlist)
    sys.exit(1)

if not os.path.exists(resultfolder):
    os.mkdir(resultfolder)

destfolder = os.path.join(resultfolder, hostname)
if not os.path.exists(destfolder):
    os.mkdir(destfolder)

write_topology(destfolder)

avail_tests = get_avail_tests()

if not only_wiki:

    scriptfile = os.path.join(destfolder, scriptfilename)
    script = open(scriptfile,'w')
    script.write("#!/bin/bash\n")


    for group in test_set.keys():
        perfctr_string = "%s -f -C E:N:%d:1:2 -g %s -m " % (perfctr, nrThreads, group,)
        no_scale = False
        for test in test_set[group].keys():
            if test.startswith("REGEX"): continue
            if test not in avail_tests:
                print "Skip Group %s Test %s" % (group, test,)
                continue
            filebase = "%s_%s" % (group, test)

            file_plain = os.path.join(destfolder,filebase+"_plain.dat")
            #raw_plain = os.path.join(destfolder,filebase+"_plain.raw")
            file_correct = os.path.join(destfolder,filebase+"_correct.dat")
            file_marker = os.path.join(destfolder,filebase+"_marker.dat")
            raw_marker = os.path.join(destfolder,filebase+"_marker.raw")
            fp_ref = open(file_plain,'w')
            fp_ref.write("#Counter Size Reference Scaled_Reference\n")
            fp_scale_ref = open(file_correct,'w')
            fp_scale_ref.write("#Counter Size Scaled_Reference\n")
            fp_likwid = open(file_marker,'w')
            fp_likwid.write("#Counter Size Measurement\n")
            rawfp_likwid = open(raw_marker,'w')
            if papi:
                file_papi = os.path.join(destfolder, filebase+"_papi.dat")
                raw_papi = os.path.join(destfolder, filebase+"_papi.raw")
                fp_papi = open(file_papi,'w')
                fp_papi.write("#Counter Size Measurement\n")
                rawfp_papi = open(raw_papi,'w')
            else:
                file_papi = None
                raw_papi = None
            counter = 1
            print "Group %s Test %s" % (group, test,)
            for size in test_set[group][test]["variants"]:
                if size.startswith("RUNS"): continue
                print "Size "+size+": ",
                bench_options = "-t %s -w N:%s:%d" % (test, size, nrThreads)
                for i in range(0,test_set[group][test]["RUNS"]):
                    print "*",
                    sys.stdout.flush()
                    # Run with LIKWID instrumented likwid-bench and likwid-perfctr
                    rawfp_likwid.write(perfctr_string+" "+bench_marker+" "+bench_options+"\n")
                    p = subprocess.Popen(perfctr_string+" "+bench_marker+" "+bench_options, shell=True, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,executable="/bin/bash")
                    stdout = ""
                    try:
                        p.wait()
                        stdout = p.stdout.read()
                        p.stdout.close()
                    except:
                        sys.exit(1)
                    found_bench = False
                    found_perfctr = False
                    for line in stdout.split("\n"):
                        if p.returncode == 0:
                            match = test_set[group]["REGEX_PERF"].match(line)
                            if match:
                                marker_set[group][test][size].append(float(match.group(1)))
                                #fp_likwid.write(str(counter)+" "+str(float(match.group(1)))+"\n")
                                fp_likwid.write("%d %d %f\n" % (counter, get_bytes(size), float(match.group(1))))
                                found_perfctr = True
                            match = test_set[group]["REGEX_BENCH"].match(line)
                            if match:
                                found_bench = True
                                value = float(match.group(1)) * test_set[group][test]["WA_FACTOR"]
                                plain_set[group][test][size].append(match.group(1))
                                corrected_set[group][test][size].append(str(value))
                                fp_ref.write("%d %d %f %f\n" % (counter, get_bytes(size), float(match.group(1)), value))
                                #fp_ref.write(str(counter)+" "+match.group(1)+"\n")
                                #fp_scale_ref.write(str(counter)+" "+str(value)+"\n")
                                fp_scale_ref.write("%d %d %f\n" % (counter, get_bytes(size), value))
                        rawfp_likwid.write(line+"\n")
                    if not found_bench:
                        value = "0"
                        plain_set[group][test][size].append(value)
                        corrected_set[group][test][size].append(value)
                        #fp_ref.write(str(counter)+" "+value+"\n")
                        fp_ref.write("%d %d %f %f\n" % (counter, get_bytes(size), float(match.group(1)), float(value)))
                        #fp_scale_ref.write(str(counter)+" "+value+"\n")
                        fp_scale_ref.write("%d %d %f\n" % (counter, get_bytes(size), value))
                        no_scale = True
                    if not found_perfctr:
                        marker_set[group][test][size].append(0)
                        #fp_likwid.write(str(counter)+" "+str(0)+"\n")
                        fp_likwid.write("%d %d %f\n" % (counter, get_bytes(size), 0.0))
                    counter += 1
                print("")
            fp_ref.close()
            fp_scale_ref.close()
            fp_likwid.close()
            rawfp_likwid.close()
            if papi:
                fp_papi.close()
                rawfp_papi.close()
            if no_scale:
                test_set[group][test]["WA_FACTOR"] = 0.0
            if out_pgf:
                pgf_file = write_pgf(group, test, file_plain, file_marker, test_set[group][test]["WA_FACTOR"],file_papi, script=script)
            if out_gnuplot: plot_file = write_gnuplot(group, test, file_plain, file_marker, test_set[group][test]["WA_FACTOR"], file_papi, script=script)
            if out_grace: grace_file = write_grace(group, test, file_plain, file_correct, file_marker, file_papi, script=script)


    script.close()
    os.chmod(scriptfile, stat.S_IRWXU)



#!/usr/bin/env python3
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Generates an agglomerated header out of various other headers.
"""

import argparse
import sys

from pathlib import Path


class Header:
  def __init__(self, include_path, open_path):
    self.path = include_path
    self.fulltext = open_path.read_text()

    lines = []
    self.includes = []

    has_seen_licence = False
    ifdef_guard = None
    for line in self.fulltext.split('\n'):
      if not has_seen_licence and line.startswith('//'):
        continue
      has_seen_licence = True

      if ifdef_guard is None and line.startswith('#ifndef'):
        ifdef_guard = line[len('#ifndef'):].strip()
      elif (ifdef_guard is not None and
            (line == f'#define {ifdef_guard}' or
             line == f'#endif  // {ifdef_guard}')):
        pass
      elif line.startswith('#include'):
        self.includes.append(line[len('#include'):].strip())
      else:
        lines.append(line)

    self.text = '\n'.join(lines).strip()


def main():
  parser = argparse.ArgumentParser(description='agglomerate some headers')
  parser.add_argument(
    '--guard',
    type=str,
    default='CWISSTABLE_H_',
    help='include guard name for the agglomerated header'
  )
  parser.add_argument(
    '--out',
    type=argparse.FileType('w', encoding='utf-8'),
    default='cwisstable.h',
    help='location to output the file to'
  )
  parser.add_argument(
    '--include_dir',
    type=Path,
    default='.',
    help='directory path that "cwisstable/*.h" includes are relative to'
  )
  parser.add_argument(
    'hdrs',
    type=Path,
    default=[Path('cwisstable/policy.h'), Path('cwisstable/declare.h')],
    nargs='*',
    help='headers to agglomerate'
  )
  args = parser.parse_args()

  hdrs = {}
  include_dir = args.include_dir.resolve()
  for open_path in args.hdrs:
    include_path = (Path.cwd() / open_path).relative_to(include_dir)
    hdrs[include_path] = Header(include_path, open_path)

  while True:
    new_hdrs = []
    for hdr in hdrs.values():
      for inc in hdr.includes:
        include_path = Path(inc.strip('"'))
        open_path = include_dir / include_path

        if include_path not in hdrs and open_path.exists():
          new_hdrs.append(Header(include_path, open_path))
    if not new_hdrs:
      break
    for hdr in new_hdrs:
      hdrs[hdr.path] = hdr

  hdr_names = {f'"{name}"' for name, _ in hdrs.items()}
  external_includes = set()
  internal_includes = set()
  for hdr in hdrs.values():
    for inc in hdr.includes:
      if inc in hdr_names:
        internal_includes.add(inc.strip('"'))
      else:
        external_includes.add(inc)

  roots = [hdr for hdr in hdrs.values() if
           str(hdr.path) not in internal_includes]
  # Order the roots predictably; this script must be idempotent
  roots.sort(key=lambda h: h.path)

  if not roots:
    print("dependency cycle detected")
    return 1

  # Toposort the include dependencies.
  unsorted = dict(hdrs)
  sorted_hdrs = []
  while roots:
    hdr = roots.pop()
    sorted_hdrs.append(hdr)
    del unsorted[hdr.path]

    new_roots = []
    for inc in hdr.includes:
      if inc not in hdr_names:
        continue

      for other_hdr in unsorted.values():
        if inc in other_hdr.includes:
          break
      else:
        new_roots.append(hdrs[Path(inc.strip('"'))])
    new_roots.sort(key=lambda h: h.path)
    roots.extend(new_roots)
  sorted_hdrs.reverse()

  o = args.out

  # Add the license, cribbing from this file itself.
  for line in Path(__file__).read_text().split('\n')[1:]:
    if not line.startswith('#'):
      break
    o.write(line.replace('#', '//') + "\n")
  o.write("\n")

  o.write("// THIS IS A GENERATED FILE! DO NOT EDIT DIRECTLY!\n")
  o.write("// Generated using unify.py, by concatenating, in order:\n")
  for hdr in sorted_hdrs:
    o.write(f'// #include "{hdr.path}"\n')
  o.write("\n")

  # Add the include guards.
  o.write(f"#ifndef {args.guard}\n")
  o.write(f"#define {args.guard}\n")
  o.write("\n")

  # Add the external includes.
  for inc in sorted(external_includes):
    o.write(f"#include {inc}\n")
  o.write("\n")

  # Add each header.
  for hdr in sorted_hdrs:
    name = f"/// {hdr.path} /"
    name += '/' * (80 - len(name))
    o.write('\n'.join([name, hdr.text, name]) + "\n\n")

  # Add the include guard tail.
  o.write(f"#endif  // {args.guard}\n")


if __name__ == '__main__': sys.exit(main() or 0)
